""" Release custom cache functions, based on `cachetools` library. """

import functools
import inspect
from typing import Any, Callable, Union

import cachetools

from ..helpers import isCoroutineFunction

__all__ = ("cache",)


def _asyncFunctionWrapperFactory(cacheObject: Any, keyGen: Callable) -> Callable:
    """
    Async decorator to wrap a function with a memorizing
    callable that saves results in a cache

    Args:
        cacheObject: cache instance
        keyGen: cache key generator

    Returns:
        wrapper
    """

    def decorator(function):
        async def _asyncFunctionWrapper(*args, **kwargs):
            key = keyGen(*args, **kwargs)
            try:
                return cacheObject[key]
            except KeyError:
                pass  # key not found
            value = await function(*args, **kwargs)
            try:
                cacheObject[key] = value
            except ValueError:
                pass  # value too large
            return value

        return _asyncFunctionWrapper

    return decorator


def _asyncMethodWrapperFactory(cacheGetter, keyGen):
    """
    Async decorator to wrap a class or instance method with a memorizing
    callable that saves results in a cache
    Args:
        cacheGetter: cache getter
        keyGen: cache key generator

    Returns:
        wrapper
    """

    def decorator(method):
        async def _asyncMethodWrapper(self, *args, **kwargs):
            cacheObject = cacheGetter(self)
            key = keyGen(*args, **kwargs)
            try:
                return cacheObject[key]
            except KeyError:
                pass  # key not found
            value = await method(self, *args, **kwargs)
            try:
                cacheObject[key] = value
            except ValueError:
                pass  # value too large
            return value

        return _asyncMethodWrapper

    return decorator


def cache(
    cacheGetter: Union[Callable[[], cachetools.Cache]],
    keyGen: Callable = cachetools.keys.hashkey,
):
    """
    Created from the `cachetools`.

    Decorator to wrap a class/instance sync/async method or function with a memoizing callable that saves results in
    the cache.

    Args:
        cacheGetter: Callable, that returns cache object when called.
            For methods is called with one "self" argument.
            For functions is called without arguments.
        keyGen: key generator from "*args", "**kwargs" from the decorating function

    See Also:
        https://cachetools.readthedocs.io/en/stable/
    """

    def decorator(function):
        isMethod = next(iter(inspect.getfullargspec(function).args), None) == "self"
        if isCoroutineFunction(function):
            if isMethod:
                wrapper = _asyncMethodWrapperFactory(cacheGetter=cacheGetter, keyGen=keyGen)(method=function)
            else:
                wrapper = _asyncFunctionWrapperFactory(cacheObject=cacheGetter(), keyGen=keyGen)(function=function)
        else:
            if isMethod:
                wrapper = cachetools.cachedmethod(cache=cacheGetter, key=keyGen, lock=None)(method=function)
            else:
                wrapper = cachetools.cached(cache=cacheGetter(), key=keyGen, lock=None)(func=function)
        return functools.update_wrapper(wrapper, function)  # I really need it! pylint: disable-msg=W0212

    return decorator
