"""
Module release functions to work with descriptors.
"""

from math import exp, inf, log, sqrt

from .data import DescriptorsEnum, PlattScaling

try:
    from math import dist  # type: ignore
except ImportError:
    dist = None  # type: ignore

ZERO_DISTANCE_STUB = 0.001


def _getPlattScaling(descriptorVersion: int | None) -> PlattScaling:
    """
    Get platt coefficients from `Descriptors` object.

    Args:
        descriptorVersion: descriptor version

    Returns:
        platt scaling coefficients
    """
    try:
        plattScaling = DescriptorsEnum[f"d{descriptorVersion}"].value.plattScaling
    except KeyError as e:
        msg = f"Fill up {DescriptorsEnum} object from '{DescriptorsEnum.__module__}' with object {e}"
        raise NotImplementedError(msg) from e
    return plattScaling


def distanceToSimilarity(
    distance: float, descriptorVersion: int | None = None, plattScaling: PlattScaling | None = None
) -> float:
    """
    Cast distance to similarity.

    Args:
        distance: distance between descriptors
        descriptorVersion: descriptor version instead of plattScalling
        plattScaling: platt scaling coef instead descriptorVersion

    Returns:
        descriptors similarity
    """
    plattScaling = plattScaling or _getPlattScaling(descriptorVersion)

    if distance <= ZERO_DISTANCE_STUB:
        return 1.0
    return 1.0 / (1.0 + exp(plattScaling.a * distance + plattScaling.b))


def similarityToDistance(
    similarity: float, descriptorVersion: int | None = None, plattScaling: PlattScaling | None = None
) -> float:
    """
    Cast similarity to distance.

    Args:
        similarity: distance between descriptors
        descriptorVersion: descriptor version instead of plattScalling
        plattScaling: platt scaling coef instead descriptorVersion

    Returns:
        descriptors distance

    >>> sims = (0.0, 0.5, 1.0)
    >>> for sim in sims:
    ...     for dv in (46, 56):
    ...         print(sim, distanceToSimilarity(similarityToDistance(sim, dv), dv))
    0.0 0.0
    0.0 0.0
    0.5 0.5
    0.5 0.5
    1.0 1.0
    1.0 1.0
    """
    if similarity == 1.0:
        return ZERO_DISTANCE_STUB
    plattScaling = plattScaling or _getPlattScaling(descriptorVersion)
    if similarity == 0.0:
        return inf
    return (log(1.0 / similarity - 1.0) - plattScaling.b) / plattScaling.a


def match(descriptor1: bytes, descriptor2: bytes, version: int) -> float:
    """
    Match two descriptors.
    !WARNING! Better use FaceMatcher object from the `FaceEngine` library!

    Args:
        descriptor1: first descriptor
        descriptor2: second descriptor
        version: descriptor version to use

    Returns:
        similarity
    >>> match(b"\x80\x7e" * 256, b"\x7e\x80" * 256, 54)
    0.9999868465292343
    """
    if len(descriptor1) != len(descriptor2):
        raise ValueError(f"Descriptors lengths differ: {len(descriptor1)} {len(descriptor2)}")
    if dist is None:
        distance = sqrt(sum((px - qx) ** 2.0 for px, qx in zip(descriptor1, descriptor2))) / 128
    else:
        distance = dist(descriptor1, descriptor2) / 128
    sim = distanceToSimilarity(distance, descriptorVersion=version)
    return sim


def matchDistance(descriptor1: bytes, descriptor2: bytes) -> float:
    """
    Match two descriptors.
    !WARNING! Better use FaceMatcher object from the `FaceEngine` library!

    Args:
        descriptor1: first descriptor
        descriptor2: second descriptor

    Returns:
        distance
    >>> match(b"\x80\x7e" * 256, b"\x7e\x80" * 256, 54)
    0.9999868465292343
    """
    if len(descriptor1) != len(descriptor2):
        raise ValueError(f"Descriptors lengths differ: {len(descriptor1)} {len(descriptor2)}")

    if dist is None:
        distance = sqrt(sum((px - qx) ** 2.0 for px, qx in zip(descriptor1, descriptor2))) / 128
    else:
        distance = dist(descriptor1, descriptor2) / 128

    return distance
