"""
XPK file reader.

Taken from http:
git.visionlabs.ru/face-recognition/PhotoAuditor/blob/07dd4ef787f5165cbf16ca673c03ef1957446524/src/xpkreader/xpkreader.py
"""

__author__ = "VisionLabs"

import struct
import sys
from typing import Dict, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

# (Dict[int, str]): map structure id to structure name
STRUCTURE_ID_TO_NAME_MAP = {3: "detection", 1: "featureSet", 0: "descriptor", 2: "portrait_detection", 4: "metadata"}
# (Dict[int, str]): map element type to element name
ELEMENT_ID_TO_NAME_MAP = {
    0: "Landmarks5",
    1: "Landmarks68",
    2: "EyesEstimation",
    3: "QualityEstimation",
    4: "AttributeEstimation",
    5: "Descriptor",
    6: "PortraitSpaceDetection",
    7: "MetaData",
    8: "GarbageScore",
    9: "SmileEstimation",
    10: "HeadPoseEstimation",
    11: "GazeEstimation",
    12: "EmotionsEstimation",
}
# (Dict[int, str]): map point index to point name
POINTS_ID_TO_NAME_MAP = {
    0: "left_eye_left_corner",
    1: "left_eye_right_corner",
    2: "right_eye_left_corner",
    3: "right_eye_right_corner",
    4: "nose_left_wing",
    5: "nose_tip",
    6: "nose_right_wing",
    7: "mouth_left_corner",
    8: "mouth_right_corner",
    9: "left_eye_center",
    10: "right_eye_center",
    11: "count",
}

# (Dict[int, str]): map landmarks5 point index to point name
LANDMARKS5_INDEX_TO_NAME_MAP = {
    0: "LeftEye",
    1: "RightEye",
    2: "Nose",
    3: "MouthLeftCorner",
    4: "MouthRightCorner",
    5: "Count",
}

# (Dict[int, str]): map landmarks5 point index to point name
EYE_STATE_ID_TO_NAME_MAP = {0: "Closed", 1: "Open", 2: "Occluded"}


def _readFeatureSet(xpkBytes: bytes) -> dict:
    """
    Read feature set section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with feature set
    """
    features = {
        "signature": struct.unpack("i", xpkBytes[:4])[0],
        "version": struct.unpack("i", xpkBytes[4:8])[0],
        "size": xpkBytes[8],
    }
    points = []
    for j in range(features["size"]):
        pointBatch = xpkBytes[9 + j * 5 : 9 + (j + 1) * 5]
        point = {
            "type": POINTS_ID_TO_NAME_MAP[pointBatch[0]],
            "x": int.from_bytes(pointBatch[1:3], byteorder="little"),
            "y": int.from_bytes(pointBatch[3:5], byteorder="little"),
        }
        points.append(point)
    features["points"] = points
    return features


class _Descriptor(TypedDict):
    signature: int  # descriptor signature
    version: int  # descriptor version
    raw_descriptor: bytes  # descriptor bytes


def readDescriptor(xpkBytes: bytes) -> _Descriptor:
    """
    Read descriptor section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with descriptor
    """
    descriptor = _Descriptor(
        signature=int.from_bytes(xpkBytes[:4], byteorder="little"),
        version=int.from_bytes(xpkBytes[4:8], byteorder="little"),
        raw_descriptor=xpkBytes[8:],
    )
    return descriptor


def _readDetection(xpkBytes: bytes) -> Dict[str, Union[float, int]]:
    """
    Read detection section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with detection
    """
    return {
        "x": int.from_bytes(xpkBytes[:4], byteorder="little", signed=True),
        "y": int.from_bytes(xpkBytes[4:8], byteorder="little", signed=True),
        "width": int.from_bytes(xpkBytes[8:12], byteorder="little", signed=True),
        "height": int.from_bytes(xpkBytes[12:16], byteorder="little", signed=True),
        "score": struct.unpack("f", xpkBytes[16:20])[0],
    }


def _readGarbageScore(xpkBytes: bytes) -> Dict[str, float]:
    """
    Read garbage score section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with garbage score
    """
    return {"score": struct.unpack("f", xpkBytes[:4])[0]}


def _readSmileEstimation(xpkBytes: bytes) -> Dict[str, float]:
    """
    Read smile section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with smile
    """
    return {
        "mouth": struct.unpack("<f", xpkBytes[:4])[0],
        "smile": struct.unpack("<f", xpkBytes[4:8])[0],
        "overlap": struct.unpack("<f", xpkBytes[8:12])[0],
    }


def _readHeadPoseEstimation(xpkBytes: bytes) -> Dict[str, float]:
    """
    Read head pose section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with head pose
    """
    return {
        "pitch": struct.unpack("<f", xpkBytes[:4])[0],
        "yaw": struct.unpack("<f", xpkBytes[4:8])[0],
        "roll": struct.unpack("<f", xpkBytes[8:12])[0],
    }


def _readGazeEstimation(xpkBytes: bytes) -> Dict[str, float]:
    """
    Read gaze estimation section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with gaze estimation
    """
    return {
        "left_yaw": struct.unpack("<f", xpkBytes[:4])[0],
        "left_pitch": struct.unpack("<f", xpkBytes[4:8])[0],
        "right_yaw": struct.unpack("<f", xpkBytes[8:12])[0],
        "right_pitch": struct.unpack("<f", xpkBytes[12:16])[0],
    }


def _readLandmarks5(xpkBytes: bytes) -> Dict[str, Dict[str, float]]:
    """
    Read landmarks5 section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with landmarks5
    """
    landmarks5 = {}
    for i in range(0, 5):
        landmarkBatch = xpkBytes[i * 8 : (i + 1) * 8]
        landmarks5[LANDMARKS5_INDEX_TO_NAME_MAP[i]] = {
            "x": struct.unpack("<f", landmarkBatch[:4])[0],
            "y": struct.unpack("<f", landmarkBatch[4:])[0],
        }
    return landmarks5


def _readLandmarks68(xpkBytes: bytes) -> Dict[int, Dict[str, float]]:
    """
    Read landmarks68 section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with landmarks68
    """
    landmarks68 = {}
    for i in range(0, 68):
        landmarkBatch = xpkBytes[i * 8 : (i + 1) * 8]
        landmarks68[i] = {
            "x": struct.unpack("<f", landmarkBatch[:4])[0],
            "y": struct.unpack("<f", landmarkBatch[4:])[0],
        }
    return landmarks68


def _readEyesEstimation(xpkBytes: bytes) -> Dict[str, dict]:
    """
    Read eyes estimation section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with eyes estimation
    """
    eyesEstimation = {}
    eyeList = ["left", "right"]
    for i in range(0, 2):
        eyeBatch = xpkBytes[i * 260 : (i + 1) * 260]
        state = EYE_STATE_ID_TO_NAME_MAP[eyeBatch[0]]
        pupilLandmarks = {}
        for j in range(0, 32):
            pupilLandmarkBatch = eyeBatch[4 + j * 8 : 4 + (j + 1) * 8]
            pupilLandmarks[j] = {
                "x": struct.unpack("<f", pupilLandmarkBatch[:4])[0],
                "y": struct.unpack("<f", pupilLandmarkBatch[4:])[0],
            }
        eyesEstimation[eyeList[i]] = {"state": state, "landmarks": pupilLandmarks}
    return eyesEstimation


def _readQualityEstimation(qualityBytes: bytes):
    """
    Read qualities section
    Args:
        qualityBytes: reading bytes

    Returns:
        dict with qualities
    """
    return {
        "light": struct.unpack("<f", qualityBytes[:4])[0],
        "dark": struct.unpack("<f", qualityBytes[4:8])[0],
        "gray": struct.unpack("<f", qualityBytes[8:12])[0],
        "blur": struct.unpack("<f", qualityBytes[12:16])[0],
        "quality": struct.unpack("<f", qualityBytes[16:20])[0],
    }


def _readAttributeEstimation(attributesBytes: bytes) -> Dict[str, float]:
    """
    Read Attribute section
    Args:
        attributesBytes: reading bytes

    Returns:
        dict with attribute
    """
    return {
        "gender": struct.unpack("<f", attributesBytes[:4])[0],
        "glasses": struct.unpack("<f", attributesBytes[4:8])[0],
        "age": struct.unpack("<f", attributesBytes[8:12])[0],
    }


def _readEmotionsEstimation(xpkBytes: bytes) -> Dict[str, float]:
    """
    Read emotion section
    Args:
        xpkBytes: reading bytes

    Returns:
        dict with emotions
    """
    return {
        "anger": struct.unpack("<f", xpkBytes[:4])[0],
        "disgust": struct.unpack("<f", xpkBytes[4:8])[0],
        "fear": struct.unpack("<f", xpkBytes[8:12])[0],
        "happiness": struct.unpack("<f", xpkBytes[12:16])[0],
        "sadness": struct.unpack("<f", xpkBytes[16:20])[0],
        "surprise": struct.unpack("<f", xpkBytes[20:24])[0],
        "neutral": struct.unpack("<f", xpkBytes[24:28])[0],
    }


def _readXPKFile(xpkBytes: bytes) -> dict:
    """
    XPK file reader function.

    XPK consists of three parts:

    Header: |XPKF|4-bytes version|4-bytes count of fields|
    Table: table of fields, every row consists of |4-bytes field name|4-bytes field data offset|4-bytes field data size|
    Content: binary data to be read according to the Table

    Args:
        xpkBytes: XPK file as bytes.

    Returns:
        Parsed XPK as dict.
    """
    # read header
    version = struct.unpack("i", xpkBytes[4:8])[0]
    countElementInTable = struct.unpack("i", xpkBytes[8:12])[0]

    # read table
    table = []
    tableBytes = xpkBytes[12 : 12 + countElementInTable * 12]
    for rowNum in range(countElementInTable):
        row = tableBytes[12 * rowNum : 12 * (rowNum + 1)]
        elementType = struct.unpack("i", row[:4])[0]
        elementOffset = struct.unpack("i", row[4:8])[0]
        elementSize = struct.unpack("i", row[8:])[0]
        table.append([elementType, elementOffset, elementSize])

    # read content
    if version == 120:
        readersMapping = {
            "Landmarks5": _readLandmarks5,
            "Landmarks68": _readLandmarks68,
            "EyesEstimation": _readEyesEstimation,
            "QualityEstimation": _readQualityEstimation,
            "AttributeEstimation": _readAttributeEstimation,
            "Descriptor": readDescriptor,
            "PortraitSpaceDetection": _readDetection,
            "MetaData": lambda b_: b_.decode("utf-8"),
            "GarbageScore": _readGarbageScore,
            "SmileEstimation": _readSmileEstimation,
            "HeadPoseEstimation": _readHeadPoseEstimation,
            "GazeEstimation": _readGazeEstimation,
            "EmotionsEstimation": _readEmotionsEstimation,
        }
    elif version == 110:
        readersMapping = {
            "detection": _readDetection,
            "portrait_detection": _readDetection,
            "featureSet": _readFeatureSet,
            "Descriptor": readDescriptor,
        }
    else:
        raise NameError(f"Version {version} is unsupported")

    resParser = {}
    for i in range(0, countElementInTable):
        elementType, elementOffset, elementSize = table[i]
        elementType = ELEMENT_ID_TO_NAME_MAP[elementType]
        if elementType in readersMapping:
            resParser[elementType] = readersMapping[elementType](xpkBytes[elementOffset : elementOffset + elementSize])

    return resParser


def readXPKFromBinary(xpkBytes: bytes) -> dict:
    """
    Read xpk as bytes.

    >>> with open('../../tests/data/descriptor.xpk', 'rb') as f:
    ...     readXPKFromBinary(f.read())

    Args:
        xpkBytes: xpk as bytes.
    Returns:
        default input json
    Raises:
        ValueError: if read is failed
    """
    try:
        return _readXPKFile(xpkBytes)
    except Exception as e:
        raise ValueError("failed parse xpk") from e
