""" Module release recursive functions to work with objects. """

from typing import Any, List, Union


def goDeep(rootObject: Union[list, dict, Any], path: List[Union[str, int]]) -> Any:
    """
    Go to an element from a dict-list-object multilevel structure.

    Args:
        rootObject: dict or list or object
        path: path to the element

    Returns:
        None if the element was not found
        element otherwise

    >>> class A:
    ...    x = {1: {'a': [1, 2, 3]}}
    >>> goDeep(A, ['x', 1, 'a', 1])
    2
    >>> goDeep(A, ['y', 1, 'a', 1])

    """
    if not path:
        return rootObject

    if isinstance(rootObject, (dict, list)):
        try:
            return goDeep(rootObject[path[0]], path[1:])  # type: ignore
        except (KeyError, IndexError, TypeError):
            return None

    if not isinstance(path[0], str):
        return None
    try:
        return goDeep(getattr(rootObject, path[0]), path[1:])
    except (TypeError, AttributeError):
        return None


def rmField(rootObject: Union[list, dict], fieldForRemoval: str) -> Union[list, dict]:
    """
    Find and remove an element from a dict-list multilevel structure by name.

    Args:
        rootObject: dict or list or object
        fieldForRemoval: field name to remove

    Returns:
        changed rootObject

    >>> x = {'foo': 'bar', 'los': {'cos': {'foo': 'bon'}}, 'can': [1, {'dos': 2, 'foo': 0xf00, 'uno': 1}, 'foo']}
    >>> rmField(x, 'foo')
    {'los': {'cos': {}}, 'can': [1, {'dos': 2, 'uno': 1}]}
    """
    if isinstance(rootObject, dict):
        rootObject.pop(fieldForRemoval, 0)
        for nextObject in rootObject.values():
            rmField(nextObject, fieldForRemoval)
    elif isinstance(rootObject, list):
        if fieldForRemoval in rootObject:
            rootObject.remove(fieldForRemoval)
        for nextObject in rootObject:
            rmField(nextObject, fieldForRemoval)
    return rootObject
