"""
Realization dataclasses with slots
"""

import dataclasses
from dataclasses import dataclass as originDataclass
from typing import Callable


def dataclass(withSlots=False) -> Callable:
    """
    Dataclasses with dict

    Need to create a new class, since we can't set __slots__
    after a class has been created.

    Make sure __slots__ isn't already set.
    """

    def slotsDataclass(cls):
        cls = originDataclass(cls)
        if "__slots__" in cls.__dict__:
            raise TypeError(f"{cls.__name__} already specifies __slots__")

        # Create a new dict for our new class.
        clsDict = dict(cls.__dict__)
        fieldNames = tuple(f.name for f in dataclasses.fields(cls))
        clsDict["__slots__"] = fieldNames
        for fieldName in fieldNames:
            # Remove our attributes, if present. They'll still be
            #  available in _MARKER.
            clsDict.pop(fieldName, None)
        # Remove __dict__ itself.
        clsDict.pop("__dict__", None)
        # And finally create the class.
        qualname = getattr(cls, "__qualname__", None)
        cls = type(cls)(cls.__name__, cls.__bases__, clsDict)
        if qualname is not None:
            cls.__qualname__ = qualname
        return cls

    if not withSlots:
        return originDataclass
    return slotsDataclass
